package org.rosuda.jrs.test;

import java.io.PrintWriter;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Future;

import javax.script.Invocable;
import javax.script.ScriptEngine;
import javax.script.ScriptEngineFactory;
import javax.script.ScriptEngineManager;
import javax.script.ScriptException;

import org.rosuda.jrs.RObject;
import org.rosuda.jrs.ROpaque;
import org.rosuda.jrs.RScriptFactory;
import org.rosuda.jrs.RexpConvert;
import org.rosuda.jrs.ScriptEngineEx;
import org.rosuda.jrs.experimental.AsyncScriptEngine;

/**
 * Basic RScript tests.
 * @author Martyn Tebby
 */
public class RScriptTest {

	public static void main(String[] args) throws Exception {
		RexpConvert.useBean = true;
//		RScriptFactory.defaultHost = "localhost";
		new RScriptTest();
	}

	public RScriptTest() throws Exception {
		log.info("RScriptTest " + (RScriptFactory.defaultHost != null ? "Rserve" : "JRI"));
		ScriptEngine engine = testDiscovery();
//		testIORedirect(engine);
		testFactory(engine);
		testEval(engine);
		testFunction(engine);
		testCall(engine);
		TestInf test = testInterface(engine);
		testOpaque(engine, test);
		testBean(test);
		testBean2(test);
		testBean3(test);
		testNamedArgs(engine, test);
		testMapModel(engine);
		testAsync(engine);
		log.info("close");
		((ScriptEngineEx)engine).close();
	}

	public ScriptEngine testDiscovery() {
		log.info("testDiscovery");
		String extension = "R";
		ScriptEngineManager manager = new ScriptEngineManager();
		ScriptEngine engine = manager.getEngineByExtension(extension);
		assert engine != null;
		log.info(engine);
		return engine;
	}

	public void testIORedirect(ScriptEngine engine) throws Exception {
		PrintWriter out = new PrintWriter("/tmp/abc.out");
		engine.getContext().setWriter(out);
//		engine.getContext().setErrorWriter(out);
		Object obj = engine.eval("readline(prompt = 'Enter something: ')");
		log.info("read obj: " + obj);
	}

	public void testFactory(ScriptEngine engine) {
		log.info("testFactory");
		ScriptEngineFactory factory = engine.getFactory();
		log.info(factory);
		log.info(factory.getEngineName());
		log.info(factory.getEngineVersion());
		log.info(factory.getExtensions());
		log.info(factory.getMimeTypes());
		log.info(factory.getNames());
		log.info(factory.getLanguageName());
		log.info(factory.getLanguageVersion());
		log.info(factory.getParameter(ScriptEngine.NAME));
		log.info(factory.getMethodCallSyntax("obj", "method", "arg1", "arg2"));
		log.info(factory.getOutputStatement("Hello World"));
		log.info(factory.getProgram("line1", "line2"));
	}

	public void testEval(ScriptEngine engine) throws Exception {
		log.info("testEval");
		Object ret = engine.eval("a = 3; print(a); a; b = 'dfe'; a");
		assert ret.equals(new Double(3));
		log.info(ret);
	}

	public void testFunction(ScriptEngine engine) throws Exception {
		log.info("testFunction");
		engine.eval("func1 = function(i, s = 'abc', d = 1, zzz = 0)" +
				"{ print(i); print(s); print(zzz); if(zzz == 0) i + 2 else zzz; }");
		engine.eval("func2 = function(i, d=1.2, s=NULL, e='NO', b=FALSE)\n" +
				"{ ret = list(i=i, d=d, s=s, e=e, b=b); print(ret); ret; }");
		engine.eval("func5 = function(a, b=1) { print(a); a }");
		int i = 2;
		String s = "str";
		double d = 3.3;
		String zzz = "abc";
		Invocable invocable = (Invocable) engine;
		Object ret = invocable.invokeFunction("func1", i, s);
		assert ret.equals(new Double(4));
		log.info(ret);
		ret = invocable.invokeFunction("func1", i, s, d, zzz);
		assert(ret.equals(zzz));
		log.info(ret);
	}

	public void testCall(ScriptEngine engine) throws ScriptException {
		log.info("testCall");
		Object ret = ((ScriptEngineEx)engine).call(null, null, "func2", 1);
		@SuppressWarnings("rawtypes")
		Map map = (Map) ret;
		assert(map.get("i").equals(new Integer(1)));
		assert(map.get("e").equals("NO"));
		assert(map.get("d").equals(new Double(1.2)));
		assert(map.get("s") == null);
		assert(map.get("b") == Boolean.FALSE);
		log.info(ret);
	}

	public TestInf testInterface(ScriptEngine engine) throws Exception {
		log.info("testInterface");
		int i = 2;
		String s = "str";
		Invocable invocable = (Invocable) engine;
		TestInf test = invocable.getInterface(TestInf.class);
		log.info(test);
		i = test.func1(i, s);
		assert i == 4;
		log.info(i);
		return test;
	}

	public void testOpaque(ScriptEngine engine, TestInf test) throws Exception {
		log.info("testOpaque");
		engine.eval("test3=function(i) { i; }");
		engine.eval("test4=function(i, e) { print(i); print(e); i; }");
		ROpaque oo = test.test3(88);
		log.info(oo);
		int i = test.test4(oo, MyEnum.YES);
		assert i == 88;
		log.info(i);
	}

	public void testBean(TestInf test) {
		log.info("testBean");
		TestBean bean = test.func2(1, 2, "3");
		assert bean.i == 1;
		assert bean.d == 2;
		assert bean.s.equals("3");
		assert bean.e == MyEnum.NO;
		assert bean.b == false;
		log.info(bean);
	}

	public void testBean2(TestInf test) {
		log.info("testBean2");
		TestBean2 bean2 = test.func2(1, 2, "3", MyEnum.YES, new boolean[] {true, false});
		assert bean2.i == 1;
		assert bean2.d == 2;
		assert bean2.s.equals("3");
		assert bean2.e == MyEnum.YES;
		assert bean2.b == true;
		log.info(bean2);

		if(true) {
			Map<?, ?> map = test.func2(4);
			assert map.get("i").equals(new Integer(4));
			log.info(map);
		}
	}

	public void testBean3(TestInf test) {
		log.info("testBean3");
		RexpConvert.useBean = false;
		TestBean3 a = new TestBean3(1);
		TestBean3 ret = test.func5(a);
		assert ret.equals(a);

		if(true) {
			Object obj = test.func5(a, 1);
			log.info(obj);
		}
	}

	public void testNamedArgs(ScriptEngine engine, TestInf test) {
		log.info("testNamedArgs");
		Object zzz = "xyz";
		engine.put("zzz", zzz);
		String ret = test.func1(4, "def", 1.1);
		assert ret.equals(zzz);
	}

	public void testMapModel(ScriptEngine engine) throws Exception {
		log.info("testMapModel");
		Map<String, Object> map = new HashMap<String, Object>();
		map.put("x", new Double[] { 2.2, 3.1, 4.4 });
		map.put("y", new int[] { 1, 2, 3 });
		ScriptEngineEx enginex = (ScriptEngineEx) engine;
		ROpaque model = enginex.call(ROpaque.class, null, "lm", "y ~ x", map);
		double[] ret = enginex.call(null, null, "predict", model, map);
		enginex.call(null, null, "print", map);
//		engine.eval("print(summary(" + model + "))");
		log.info(Arrays.toString(ret));
	}

	public void testAsync(ScriptEngine engine) throws Exception {
		log.info("testAsync");
		AsyncScriptEngine aengine = new AsyncScriptEngine(Collections.singletonList((ScriptEngineEx)engine));
		Object ret = aengine.invokeFunction("func1", 3, "str");
		assert ret == null;
		Future<?> future = aengine.getFuture();
		ret = future.get();
		assert ret.equals(new Double(5));
		log.info(ret);
		aengine.close();
	}

	enum MyEnum { YES, NO };

	interface TestInf {
		int func1(int i, String s);
		String func1(int i, String s, double d);
		TestBean func2(int i, double d, String s);
		TestBean2 func2(int i, double d, String s, MyEnum e, boolean[] b);
		Map<?, ?> func2(int i);
		ROpaque test3(int i);
		int test4(ROpaque o, MyEnum e);
		TestBean3 func5(TestBean3 a);
		Object func5(TestBean3 a, int b);
	}

	public static class TestBean {
		public int i;
		public double d;
		public String s;
		public MyEnum e;
		public boolean b;

		public String toString()
			{ return getClass() + " " + i + " " + d + " " + s + " " + e + " " + b; }

		public int getI()			{ return i; }
		public void setI(int i)		{ this.i = i; }
		public double getD()		{ return d; }
		public void setD(double d)	{ this.d = d; }
		public String getS()		{ return s; }
		public void setS(String s)	{ this.s = s; }
		public MyEnum getE()		{ return e; }
		public void setE(MyEnum e)	{ this.e = e; }
		public boolean getB()		{ return b; }
		public void setB(boolean b)	{ this.b = b; }
	}

	public static class TestBean2 extends TestBean implements RObject {}

	public static class TestBean3 implements RObject {
		public boolean b;	public Boolean B;		public boolean[] bs;	public Boolean[] Bs;
		public int i;		public Integer I;		public int[] is;		public Integer[] Is;
		public double d;	public Double D;		public double[] ds;		public Double[] Ds;
		public double[][] dss;
		public String str;							public String[] strs;
		public Map<String, String> map;
		public Object[] array;
		public List<Object> list;
		public MyEnum enm;
		public byte[] bytes;
		public Object Null;

		public TestBean3() {}

		public TestBean3(int a) {
			b = true;	B = false;	bs = new boolean[] { false, true };	Bs = new Boolean[] { true, false };
			i = 1;		I = 2;		is = new int[] { 3, 4 };			Is = new Integer[] { 5, 6 };
			d = 1.1;	D = 2.2;	ds = new double[] { 3.3, 4.4 };		Ds = new Double[] { 5.5, 6.6, Double.NaN };
			dss = new double[][] { { 1.1, 2.2 }, { 3.3, 4.4 } };
			str = "abc";			strs = new String[] { "def", "123" };
			map = new HashMap<String, String>();
			map.put("1", "a"); map.put("2", "b");
			array = new Object[] { 1, 2, 3, 4 };
			list = Arrays.asList(array);
			enm = MyEnum.YES;
			bytes = new byte[] { 1, 2, 3, 4 };
		}

		public boolean equals(Object that) {
			if(this == that) return true;
			if(!(that instanceof TestBean3)) return false;
			TestBean3 o = (TestBean3) that;
			return
			b == o.b && B.equals(o.B) && Arrays.equals(bs, o.bs) && Arrays.equals(Bs, o.Bs) &&
			i == o.i && I.equals(o.I) && Arrays.equals(is, o.is) && Arrays.equals(Is, o.Is) &&
			d == o.d && D.equals(o.D) && Arrays.equals(ds, o.ds) && Arrays.equals(Ds, o.Ds) &&
			Arrays.deepEquals(dss, o.dss) &&
			str.equals(o.str) && Arrays.equals(strs, o.strs) &&
			map.equals(o.map) &&
			Arrays.equals(array, o.array) &&
			list.equals(o.list) && 
			enm == o.enm &&
			Arrays.equals(bytes, o.bytes) &&
			Null == o.Null &&
			true;
		}
	}

	private static class log {
		static void info(Object str) {
			System.out.println(str);
		}
	}
}
